/*
File: CBinomial.h
Purpose: Creates a base class CBinomial
which optimizes all the necessary likelihoods
Created: 10/19/05
Creator: Matt Wheeler
Last Modefied: 
8/25/06 Added code that scaled the dose to be between 0 and 1 and then rescale 
        that dose, this allows for greater numerical stability. 
6/11/07 Modified the Hessian matrix calculation for near zero estimates
*/ 
#include <string.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_cdf.h>
#include "CBinomial.h"
#include "run_dmngb.h"

// Default constructor for the class
CBinomial::CBinomial(){
         
        nparms                   = 1;         
        error_status             = 0; // current error status
        logLikelihood            = 0.0;   // current value of the logLikelihooderror_status 
        isData                   = false;  

        model_parameters         = new(double[nparms]); 
        upper_bounds             = new(double[nparms]); 
        lower_bounds             = new(double[nparms]); 
        
        hasBeenMaximized         = false; 
        // sets all of the active bounds to zero      
         
} 


// Destructor for the class
CBinomial::~CBinomial(){
      if (model_parameters != NULL)                 
          delete model_parameters;    
      if (upper_bounds != NULL) 
          delete upper_bounds;            
      if (lower_bounds != NULL)
          delete lower_bounds;           
}

///////////////////////////////////////////////////////////////////////////
//Function:  Copy Constructor
//Purpose:   Copies the binomial class
//Input:     A binomial class
//Output:    A CBinomial class which is exactly the same as b; 
//Result:    An 
CBinomial::CBinomial(const CBinomial &b) {
         // delete from memory the current estimate
      if (model_parameters != NULL)                 
          delete model_parameters;    
      if (upper_bounds != NULL) 
          delete upper_bounds;            
      if (lower_bounds != NULL)
          delete lower_bounds;
      
         
      // get the new estimates
    
      nparms                   = b.nparms; 
      error_status             = b.error_status; // current error status
      logLikelihood            = b.logLikelihood;   // current value of the logLikelihooderror_status 
      isData                   = b.isData;  
      data                     = b.data; 
      
      model_parameters         = new(double[nparms]); 
      upper_bounds             = new(double[nparms]); 
      lower_bounds             = new(double[nparms]);     
      
     for (int i = 0; i < nparms; i++){
         model_parameters[i] = b.model_parameters[i]; 
         lower_bounds[i] = b.lower_bounds[i]; 
         upper_bounds[i] = b.upper_bounds[i];
     }     
          
}

//Function : set_numParms(int numberParms)
//Purpose:  reinitializes the class based upon
// the new number of parameters
void CBinomial::set_numParms(int numberParms){
      
      // delete from memory the current estimate
      if (model_parameters != NULL)                 
          delete model_parameters;    
      if (upper_bounds != NULL) 
          delete upper_bounds;            
      if (lower_bounds != NULL)
          delete lower_bounds;
      
         
      // get the new estimates
      nparms                   = numberParms;         
      error_status             = 0; // current error status
      logLikelihood            = 0.0;   // current value of the logLikelihooderror_status 
      isData                   = false;  

      model_parameters         = new(double[nparms]); 
      upper_bounds             = new(double[nparms]); 
      lower_bounds             = new(double[nparms]);
}

//Function : get_numParms()
//Purpose : returns the number of parameters
int CBinomial::get_numParms() {
  return nparms;
}

//Function : copy_Parms(double *)
//Purpose : copies the parameter vector to the array pointed to in
// the argument.
//void CBinomial::copy_Parms(double v[][60], int row, int strcol){
void CBinomial::copy_Parms(double *v){
  for (int i=0; i < nparms; i++) {
    v[i] = model_parameters[i];
  }
}

///////////////////////////////////////////////////////////////
// Functions: 
// setData(double*array,int row, int col)
// setData(CMatrix c)
// Purpose: 
// Updates the Classes data matrix to a new data set
void CBinomial::setData(double * array,int row, int col){
     data.InitFromArray(array,row,col); 
     isData                   = true; 
     
}

void CBinomial::setData(CMatrix c){
     data = c; 
     isData = true; 
}

void CBinomial::setUpperBounds(const double * v){   
     for (int i = 0; i < nparms; i++){
         upper_bounds[i] = v[i];
     }     
}

void CBinomial::setLowerBounds(const double * v){
     for (int i = 0; i < nparms; i++){
         lower_bounds[i] = v[i];
     }
     
} 
///////////////////////////////////////////////////////////////////////////
//Function:  operator = 
//Purpose:   Copies the binomial class
//Input:     A binomial class
//Output:    A CBinomial class which is exactly the same as b; 
//Result:    It copies the data from one class to another
//           not changing any of the optimization code
CBinomial & CBinomial::operator = (CBinomial &b) {
         // delete from memory the current estimate

      
      isData                   = b.isData;  
      data                     = b.data; 
      
      if (b.class_type == class_type) // they are the same type of optimization 
                                      // class
      {                                         
          // get the new estimates
          if (model_parameters != NULL)                 
              delete model_parameters;    
          if (upper_bounds != NULL) 
              delete upper_bounds;            
          if (lower_bounds != NULL)
              delete lower_bounds;
          nparms                   = b.nparms; 
          error_status             = b.error_status; // current error status
          logLikelihood            = b.logLikelihood;   // current value of the logLikelihooderror_status 
     
    
          model_parameters         = new(double[nparms]); 
          upper_bounds             = new(double[nparms]); 
          lower_bounds             = new(double[nparms]);     
          
         for (int i = 0; i < nparms; i++){
             model_parameters[i] = b.model_parameters[i]; 
             lower_bounds[i] = b.lower_bounds[i]; 
             upper_bounds[i] = b.upper_bounds[i];
         }     
         
       }       
     return *this; 
         
}


//////////////////////////////////////////////////////////////////////////
//Function: log_likelihood
//Input:   
//      d - a vector of doubles which coorespond to the link function
//Output: 
//      returns the negative log likelihood value; 
//////////////////////////////////////////////////////////////////////////

double CBinomial::log_likelihood(const double *d){
      
       double sum = 0.0; //set the sum of the log-likelihood to zero
       double n,obs,dose,p=0.01; 
       int nrow = data.nrow();
 
       // for each data row compute the associated 
       // negative log likelihood
       for ( int i = 0; i < nrow; i++){
           n = data.get(i,1);    obs = data.get(i,2);
           dose = data.get(i,0); 
           p = link_function(d,dose); 
           //printf("p:%f",p); 
           if ( p <= 0){
                if ( obs == 0 ) {
                     sum += 0; 
                } else{
                     sum += 1e3; 
                }
           }else if(p >= 1.0){
                if ( obs == n ) {
                     sum += 0; 
                } else{
                     sum += 1e4;                }
           }else{
                sum += -1*(obs*log(p) + (n-obs)*log(1-p)); 
           }
 
       }
       return sum;        
}
/////////////////////////////////////////////////////////////////
// Function: link_function(const double *v, double dose)
// Purpose : Defines a link function in the form of p = g(v,dose) where
//           v is a vector of parameters to the function  and dose is the 
//           specified dose
// Input   :
//          const double *v - vector of parameters to the link function
//              double dose - dose associated with a probability of response
// Output  :
//          return          - the probability of response associated with the link
// Type of link: Identity
//          p = mu 
///////////////////////////////////////////////////////////////////
double  CBinomial::link_function(const double *v,double dose){
       double mu = v[0]; 
       double p;                
       p = mu; 
       return p;      
}
/////////////////////////////////////////////////////////////////
// Function: gradient(const double *v, double *g)
// Purpose : Performs a finite difference evaluation of the 
//           gradient  
// Input   :
//          const double *v - vector of parameters to the link function
//  
// Output  :
//          double * g      - vector of gradients cooresponding to the 
//                            parameters v
// Last Modified:  11/16/05
///////////////////////////////////////////////////////////////////
void CBinomial::gradient(const double *v, double *g){
       double h[nparms]; 
       double mpres = pow(1.0e-16,0.333333);
       double x,temp;
       double derEst; 
       double f1=0.0; double f2=0.0;
       double *hvector = new(double[nparms]);   
       memcpy(hvector,v,nparms*sizeof(double));
       
       
       for(int i =  0; i < nparms; i++)
       {
          if ( x > 1.0e-7){        
              h[i] = mpres*(fabs(x)); 
              temp = x + h[i]; 
              h[i] = temp - x; 
          }else{
              h[i] = mpres;       
          }
       }
       /*find the gradients for each of the variables in the likelihood*/                       
       for (int i = 0; i < nparms; i++){
          /*perform a finite difference calculation on the specific derivative*/
          x = v[i];  

          // add h
          hvector[i] = x+h[i];  
          f1 = log_likelihood(hvector); 
          // subtract h
          hvector[i] = x-h[i];
          f2 =  log_likelihood(hvector);
           // estimate the derivative
          g[i] = (f1-f2)/(2.0*h[i]); 
          hvector[i] = x;   
      }       
  /*    printf("%f\n %f\n %f\n", v[0], v[1], v[2]);
      printf("%f\n %f\n %f\n", g[0], g[1], g[2]);
      system("PAUSE") ;     */
      
      
      delete(hvector);          
}
//////////////////////////////////////////////////////////////////////////////
// static friend functions that are necessary for the optimization using
// fortran optimizer
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//Function: binomial_g(long int *nvar, double *x, long int *nf, double *g,
//                    long int *uiparm, double *urparm, void (*ufparm) ())
// input: 
//			nvar    - is the number of parameters
//			nf      - is something for the fortran maximization--not used
//			uiparm  - is hijacked and used to figure out which class to call
//			urparm  - is a vector of fixed parameters
//			ufparm  - something for the fortran maximization--not used
//          x       - is a vector of non-fixed parameters of length nvar
//		output:          
//			g isthe gradient value
//Modified 11/16/2006
//////////////////////////////////////////////////////////////////////////////
void binomial_g(long int *nvar, double *x, long int *nf, double *g,
	   long int *uiparm, double *urparm, void (*ufparm) ())
{
     CBinomial *val = (CBinomial *)(void*) uiparm;
     val->gradient(x,g);     
}

/////////////////////////////////////////////////////////////////////////////
//Function:  binomial_lk -- used to compute the (minus) log likelihood for 
//                          a binomial likelihood
//		input:
//			nvar       - is the number of parameters
//			nf         - is something for the fortran maximization--not used
//			uiparm     - is hijacked and used to figure out which class to call
//  		urparm     - is a vector of fixed parameters
//			ufparm     - is something for the fortran maximization--not used
//          x          - is a vector of non-fixed parameters
//		output:
//			f is the likelihood value
//
//Modified 11/16/2005
//////////////////////////////////////////////////////////////////////////////

void binomial_lk(long int *nvar,  double *x, long int *nf, double *f,
	             long int *uiparm,double *urparm, void (*ufparm)()) {
        CBinomial *val = (CBinomial *)(void*) uiparm;
        *f = val->log_likelihood(x); 
}
/////////////////////////////////////////////////////////////////////////////
//Function:  max_likelihod -- called to maximize the likelihood for a 
//                          a binomial dose-response likelihood
//		input:
//			iter       - number of iterations
//			gtol       - parameter tolerance
//			relConv    - relative convergence
// Note: This maximization call will be run by most of the routines, the 
//       weibull, multistage, gamma and logistic models need to rescale the
//       doses. 
//Modified 8/25/2006
//////////////////////////////////////////////////////////////////////////////
void CBinomial::max_likelihod(int iter,double gtol,double relConv){
     int ErrorFlag;
     double *urparm = new (double[nparms]);
     double maxdose = 0.0; 
     double tempdose = maxdose; 
     if (isData){ 
        initializeParameters(); 
      
        double scales =  1;  
     
        ErrorFlag = run_dmngb(nparms,model_parameters,lower_bounds, upper_bounds,
                              scales,gtol,relConv,iter,10,
			                 (void (*)())binomial_lk,(void(*)())binomial_g,
                             (long int *) this,urparm,NULL,0,&logLikelihood);
        hasBeenMaximized = true; 
        // rescale parameters and doses
        
     }


     delete urparm;     
}

void CBinomial::print_output(){
     printf("%f\n",2*logLikelihood); 
     printf("Parm1:%f\n",model_parameters[0]); 
     printf("Parm2:%f\n",model_parameters[1]); 
     printf("Parm3:%f\n",model_parameters[2]); 
     printf("Parm4:%f\n",model_parameters[3]);
     printf("Erorr Flag:%d\n",error_status); 
     //printf("The Error is:%d\n", ErrorFlag);  
     
} 

// varMatrix
// Purpose: Computes the Variance-Covariance matrix for the 
// MLE. Estimates are derived from the observed fisher information
// Which is the inverse of the hessian matrix, evaluated at the MLE.
// The values are estimated using finite difference approximations. 
// Returns: The variance-covariance matrix of the parameters
CMatrix CBinomial::varMatrix(){
     CMatrix m; 
     m.zero(nparms);  
     double temp;  
     double h = 1e-8; 
     double mpres = pow(1.0e-16,0.333333);
     double tempVect[nparms]; 
     double hi, hj; 
     int num_nonBounded = 0; 

  
     if (true){ // if the function has been maximized
     // Compute the hessian using finite difference approximations
    
            for (int i = 0; i < nparms; i++){
              for (int j = 0; j < nparms; j++){
                 double x = model_parameters[i];   
                 if ( x > DBL_EPSILON){        
                    h = mpres*(fabs(x)); 
                    temp = x + h; 
                    hi = temp - x; 
                 }else{
                    hi = mpres;       
                 } 
                 
                 x = model_parameters[j];   
                 if ( x > DBL_EPSILON){        
                    h = mpres*(fabs(x)); 
                    temp = x + h; 
                    hj = temp - x; 
                 }else{
                    hj = mpres;       
                 } 

                 temp = 0.0; 
                 if ( lower_bounds[i] + 1e-8 < model_parameters[i] &&
                      upper_bounds[i] - 1e-8> model_parameters[i] &&
                      lower_bounds[j] + 1e-8< model_parameters[j] &&
                      upper_bounds[j] - 1e-8> model_parameters[j]){
    
                     if (i == j){
                        num_nonBounded++; 
                        // on diagonal variance estimate  
                        // eg the 2nd partial derivative
    
                        memcpy(tempVect,model_parameters,sizeof(double)*nparms); 
                        tempVect[i] = tempVect[i] + 2*hi;    
                        temp +=  -1.0*(this)->log_likelihood(tempVect);
                        memcpy(tempVect,model_parameters,sizeof(double)*nparms); 
                        tempVect[i] = tempVect[i] + hi; 
                        temp +=  16.0*(this)->log_likelihood(tempVect);
                        
                        temp += -30.0*(this)->log_likelihood(model_parameters); 
                        memcpy(tempVect,model_parameters,sizeof(double)*nparms); 
                        tempVect[i] = tempVect[i] -  hi; 
                        temp +=  16.0*(this)->log_likelihood(tempVect); 
                        memcpy(tempVect,model_parameters,sizeof(double)*nparms); 
                        tempVect[i] = tempVect[i] - 2*hi;
                        temp +=  -1.0*(this)->log_likelihood(tempVect); 
                        temp = temp/(12*hi*hi); 
                        
                        m.set(i,i,temp); 
                     }else{
                        // off diagonal variance estimate  
                        // eg the first partial of x wrt the 
                        // first partial  of y
                        memcpy(tempVect,model_parameters,sizeof(double)*nparms); 
                        tempVect[i] = tempVect[i] + hi; 
                        tempVect[j] = tempVect[j] + hj; 
                        temp +=  (this)->log_likelihood(tempVect); 
                        
                        memcpy(tempVect,model_parameters,sizeof(double)*nparms); 
                        tempVect[i] = tempVect[i] + hi; 
                        tempVect[j] = tempVect[j] - hj;                     
                        temp +=  -1.0*(this)->log_likelihood(tempVect);
                        
                        memcpy(tempVect,model_parameters,sizeof(double)*nparms); 
                        tempVect[i] = tempVect[i] - hi; 
                        tempVect[j] = tempVect[j] + hj;    
                        temp +=  -1.0*(this)->log_likelihood(tempVect);
                             
                        memcpy(tempVect,model_parameters,sizeof(double)*nparms); 
                        tempVect[i] = tempVect[i] - hi; 
                        tempVect[j] = tempVect[j] - hj;                                      
                        temp +=  (this)->log_likelihood(tempVect);  
                        temp = temp/(4*hi*hj); 
                        m.set(i,j,temp);   
                     }
                 }
             }    
         }     
         
         
        // m = m.inverse();
     }
     
     CMatrix temp_var; 
     
     if(num_nonBounded > 0){ 
         temp_var.zero(num_nonBounded); 
         bool flag_add = false; 
         int temp_i = 0 , temp_j = 0 ;      
         for (int i = 0; i < nparms; i++){
             for (int j = 0; j < nparms; j++){
                 
                   if (   lower_bounds[i] + 1e-8 < model_parameters[i] &&
                          upper_bounds[i] - 1e-8> model_parameters[i] &&
                          lower_bounds[j] + 1e-8< model_parameters[j] &&
                          upper_bounds[j] - 1e-8> model_parameters[j]){
                          temp_var.set(temp_i,temp_j,m.get(i,j));
                          temp_j++; 
                          flag_add = true;             
                     }
             }
             
             if (flag_add == true){
                temp_j = 0; 
                temp_i++; 
             }
             flag_add = false;
         }
    
         if (num_nonBounded > 0)
              temp_var = temp_var.inverse(); 
         temp_i = 0 ;  temp_j = 0;      
         for (int i = 0; i < nparms; i++){
             for (int j = 0; j < nparms; j++){
                 
                   if (   lower_bounds[i] + 1e-8 < model_parameters[i] &&
                          upper_bounds[i] - 1e-8> model_parameters[i] &&
                          lower_bounds[j] + 1e-8< model_parameters[j] &&
                          upper_bounds[j] - 1e-8> model_parameters[j]){
                          m.set(i,j,temp_var.get(temp_i,temp_j));
                          temp_j++; 
                          flag_add = true;             
                     }
             }
             
             if (flag_add == true){
                temp_j = 0; 
                temp_i++; 
             }
             flag_add = false;
         }
     }
         
     return m; 
}
