#include <gsl/gsl_multimin.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>
#include <string.h>

#include "CGamma.h"
#include "run_dmngb.h"

CGamma::CGamma(){
  set_numParms(3); 
  class_type = gamma; 
} 

// Destructor for the class
CGamma::~CGamma(){
      // do not need to free up memory as it is already done by 
      // the CBinomial Destructor   
}

//////////////////////////////////////////////////////////////////
// Function: link_function(const double *v, double dose)
// Purpose : Defines a link function in the form of p = g(v,dose) where
//           v is a vector of parameters to the function  and dose is the 
//           specified dose
// Input   :
//          const double *v - vector of parameters to the link function
//              double dose - dose associated with a probability of response
// Output  :
//          return          - the probability of response associated with the link
// Type of link: 
//          p = gamma+(1-gamma)*integral(0,beta*dose)[gamma(alpha)] 
///////////////////////////////////////////////////////////////////
double  CGamma::link_function(const double *v,double dose){
       double alpha =  v[GAMMA_ALPHA];
       double beta  =  v[GAMMA_BETA]; 
       double gamma =  v[GAMMA_GAMMA]; 
       if (alpha <= 0) alpha = 1e-8; 
      // printf("alpha:%f:beta:%f:gamma:%f\n",alpha,beta,gamma); 
        
       double p = 0.0; 
       int pos; 
       double curval; 
       
       if (dose == 0){
          p = gamma; 
       }else{
          p = gamma + (1-gamma)*gsl_cdf_gamma_P(beta*dose,alpha,1);
       }
       
       // maintain the likelihood function
       if (p< 0) p = 0; 
       if (p > 1) p = 1;
       
       return p; 
}

void CGamma::initializeParameters(){
      //initialize parameters based on a grid search over the range 
      //alpha=1..10 by 1
      //beta =0.25..2.25 by 0.5
      //gamma=0.01..0.21 by 0.5
      
      const double s_alpha = 0.25; 
      const double s_beta  = 0.50;
      const double s_gamma = 0.01; 
      double like_val = 0.0, temp_like_val=0.0; 
      double *v = new(double[nparms]);
      double *temp = new(double[nparms]);
      v[GAMMA_ALPHA] = s_alpha;
      v[GAMMA_GAMMA] = s_gamma;
      v[GAMMA_BETA]  = s_beta;
    
      like_val = log_likelihood(v); 
      memcpy(temp,v,nparms*sizeof(double)); 
       
      for (int i = 0; i < 10; i++){
          v[GAMMA_ALPHA] = s_alpha+i*0.5;          
          for (int j = 0; j < 8; j++){
              v[GAMMA_GAMMA] = s_gamma+j*0.05;
              for (int k = 0; k<12; k++){                         
                   v[GAMMA_BETA] = s_beta+k*0.2;
                   temp_like_val =  log_likelihood(v); 
                   //printf("log_like:%f",temp_like_val);             
                   if ( temp_like_val < like_val){
                      memcpy(temp,v,nparms*sizeof(double));
                      like_val = temp_like_val;       
                   }    
              }    
          }    
      }
     // gsl_vector_fprintf(stdout,temp,"%f"); 
      memcpy(model_parameters,temp,nparms*sizeof(double)); 
      delete(temp);
      delete(v); 
}
void CGamma::print_output(){
     CMatrix m; 
     m = this->varMatrix() ; 
     double temp;
     double bic = ComputeBIC((CBinomial *)this); 
     double aic = ComputeAIC((CBinomial *)this); 
     printf("MODEL: Gamma:\n"); 
     printf("------------------------------------------\n"); 
     printf("Parameters\tEstimate\tStdErr\n"); 
     printf("------------------------------------------\n");  
     temp = sqrt(m.get(2,2)); 
     if (temp == 0){
          printf("GAMMA:\t\t%f\tN/A\n",model_parameters[2]); 
     }else{
          printf("GAMMA:\t\t%f\t%f\n",model_parameters[2],temp); 
     }

     temp = sqrt(m.get(0,0)); 
     if (temp == 0){
          printf("ALPHA:\t\t%f\tN/A\n",model_parameters[0]); 
     }else{
          printf("ALPHA:\t\t%f\t%f\n",model_parameters[0],temp);            
     }     
     
     temp = sqrt(m.get(1,1)); 
     if (temp == 0){
          printf("BETA:\t\t%f\tN/A\n\n",model_parameters[1]); 
     }else{
          printf("BETA:\t\t%f\t%f\n\n",model_parameters[1],temp);            
     }           
    
     if (error_status == 0) {
          printf("Optimization Succeeded\n"); 
     } else{
          printf("****WARNING OPTIMIZATION FAILED: Results Suspect****\n"); 
     }             
} 
/////////////////////////////////////////////////////////////////////////////
//Function:  max_likelihod -- called to maximize the likelihood for a 
//                          a binomial dose-response likelihood
//		input:
//			iter       - number of iterations
//			gtol       - parameter tolerance
//			relConv    - relative convergence
// Note: This maximization call will be run by most of the routines, the 
//       weibull, multistage, gamma and logistic models need to rescale the
//       doses. 
//Modified 8/25/2006
//////////////////////////////////////////////////////////////////////////////
void CGamma::max_likelihod(int iter,double gtol,double relConv){
     int ErrorFlag;
     double *urparm = new (double[nparms]);
     double maxdose = 0.0; 
     double tempdose = maxdose; 
     if (isData){ 
        
        ///////////////////////////////////////////////////
        //scale doses to be 0<=x<=1
        //find max dose; 
        for (int i = 0; i < data.nrow(); i++){
            tempdose = data.get(i,0); 
            if (tempdose > maxdose)
                maxdose = tempdose; 
        }

        for (int i = 0; i < data.nrow(); i++){
            data.set(i,0,data.get(i,0)/maxdose);    
        }      

        //////////////////////////////////////////////////
        initializeParameters(); 
        double scales =  1;  
     
        ErrorFlag = run_dmngb(nparms,model_parameters,lower_bounds, upper_bounds,
                              0,gtol,relConv,iter,10,
			                 (void (*)())binomial_lk,(void(*)())binomial_g,
                             (long int *) this,urparm,NULL,0,&logLikelihood);
        hasBeenMaximized = true; 
        // rescale parameters and doses
        ///////////////////////////////////////////////////
        //resclae dose & the parameters to the origional scale 
        //NOTE: SPECIAL CODE FOR THE GAMMA MODEL
         for (int i = 0; i < data.nrow(); i++){
              data.set(i,0,data.get(i,0)*maxdose);    
         }    
         model_parameters[GAMMA_BETA ] = model_parameters[GAMMA_BETA ]*(1/maxdose);            
         //////////////////////////////////////////////////
        
     }
 
     delete urparm;     
}


