#include <gsl/gsl_multimin.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>
#include <string.h>
#include "CLogLogit.h"

CLogLogistic::CLogLogistic(){
  set_numParms(3); 
  class_type = log_logistic; 
} 

// Destructor for the class
CLogLogistic::~CLogLogistic(){
      // do not need to free up memory as it is already done by 
      // the CBinomial Destructor   
}

//////////////////////////////////////////////////////////////////
// Function: link_function(const double *v, double dose)
// Purpose : Defines a link function in the form of p = g(v,dose) where
//           v is a vector of parameters to the function  and dose is the 
//           specified dose
// Input   :
//          const double *v - vector of parameters to the link function
//              double dose - dose associated with a probability of response
// Output  :
//          return          - the probability of response associated with the link
// Type of link: Log-Logistic
//          p = gamma+(1-gamma)/(1+exp(-alpha+beta*log(dose)) ) 
// Modified 11/17/05
///////////////////////////////////////////////////////////////////
double  CLogLogistic::link_function(const double *v,double dose){
       double alpha =  v[LOGLOGISTIC_ALPHA];
       double beta  =  v[LOGLOGISTIC_BETA]; 
       double gamma =  v[LOGLOGISTIC_GAMMA]; 
       double p = 0.0; 
       int pos; 
       double curval; 
  
       if (dose == 0){
           p = gamma; 
       }else{
           p = gamma + (1-gamma)/(1+exp(-(alpha+beta*log(dose)))); 
       }
       // maintain the likelihood function
       if (p< 0) p = 0; 
       if (p > 1) p = 1;
//     if (p< BINOMIAL_ZERO) p = BINOMIAL_ZERO; 
//     if (p > BINOMIAL_ONE) p = BINOMIAL_ONE; 
      return p; 
}

void CLogLogistic::initializeParameters(){
     
      const double s_alpha = -1; 
      const double s_beta  = 0.25;
      const double s_gamma = 0.00; 
      double like_val = 0.0, temp_like_val=0.0; 
      double *v = new(double[nparms]);
      double *temp = new(double[nparms]);
      v[LOGLOGISTIC_ALPHA] = s_alpha;
      v[LOGLOGISTIC_GAMMA] = s_gamma;
      v[LOGLOGISTIC_BETA]  = s_beta;
    
      like_val = log_likelihood(v); 
      memcpy(temp,v,nparms*sizeof(double)); 
       
      for (int i = 0; i < 10; i++){
          v[LOGLOGISTIC_ALPHA] = s_alpha+i*0.5;          
          for (int j = 0; j < 8; j++){
              v[LOGLOGISTIC_GAMMA] = s_gamma+j*0.05;
              for (int k = 0; k<12; k++){                         
                   v[LOGLOGISTIC_BETA] = s_beta+k*0.2;
                   temp_like_val =  log_likelihood(v); 
                   //printf("log_like:%f",temp_like_val);             
                   if ( temp_like_val < like_val){
                      memcpy(temp,v,nparms*sizeof(double));
                      like_val = temp_like_val;       
                   }    
              }    
          }    
      }
     // gsl_vector_fprintf(stdout,temp,"%f"); 
      memcpy(model_parameters,temp,nparms*sizeof(double)); 
      delete(temp);
      delete(v);      
}

void CLogLogistic::print_output(){
     CMatrix m; 
     m = this->varMatrix() ; 
     double temp;
     double bic = ComputeBIC((CBinomial *)this); 
     double aic = ComputeAIC((CBinomial *)this); 
     printf("MODEL: Log-Logistic:\n"); 
     printf("------------------------------------------\n");    
     printf("Parameters\tEstimate\tStdErr\n"); 
     printf("------------------------------------------\n"); 
     temp = sqrt(m.get(2,2)); 
     if (temp == 0){
          printf("GAMMA:\t\t%f\tN/A\n",model_parameters[2]); 
     }else{
          printf("GAMMA:\t\t%f\t%f\n",model_parameters[2],temp); 
     }

     temp = sqrt(m.get(0,0)); 
     if (temp == 0){
          printf("ALPHA:\t\t%f\tN/A\n",model_parameters[0]); 
     }else{
          printf("ALPHA:\t\t%f\t%f\n",model_parameters[0],temp);            
     }     
     
     temp = sqrt(m.get(1,1)); 
     if (temp == 0){
          printf("BETA:\t\t%f\tN/A\n\n",model_parameters[1]); 
     }else{
          printf("BETA:\t\t%f\t%f\n\n",model_parameters[1],temp);            
     }
     
     if (error_status == 0) {
          printf("Optimization Succeeded\n"); 
     } else{
          printf("****WARNING OPTIMIZATION FAILED: Results Suspect****\n"); 
     }             
} 


