/*Modification List: 
               6/12/2007, Matt Wheeler, Changed the init routine to do the first
                                       iteration of a fisher scoring algorithm
*/                                       
               
               

#include <gsl/gsl_multimin.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>
#include <string.h> 
#include "CLogProbit.h"

CLogProbit::CLogProbit(){
  set_numParms(3); 
  class_type = log_probit; 
} 

// Destructor for the class
CLogProbit::~CLogProbit(){
      // do not need to free up memory as it is already done by 
      // the CBinomial Destructor   
}

//////////////////////////////////////////////////////////////////
// Function: link_function(const double *v, double dose)
// Purpose : Defines a link function in the form of p = g(v,dose) where
//           v is a vector of parameters to the function  and dose is the 
//           specified dose
// Input   :
//          const double *v - vector of parameters to the link function
//              double dose - dose associated with a probability of response
// Output  :
//          return          - the probability of response associated with the link
// Type of link: Log-Probit
//          p = gamma+(1-gamma)*(QNORM(-alpha+beta*log(dose) ) 
// Modified 11/17/05
///////////////////////////////////////////////////////////////////
double CLogProbit::link_function(const double *v,double dose){
       double alpha =  v[LOGPROBIT_ALPHA];
       double beta  =  v[LOGPROBIT_BETA]; 
       double gamma =  v[LOGPROBIT_GAMMA]; 
       double p = 0.0; 
       int pos; 
       double curval; 
       
       if (dose == 0){
           p = gamma; 
       }else{
           p = gamma+ (1-gamma)* gsl_cdf_gaussian_P(alpha+beta*log(dose),1); 
       }
      
      // printf("dose:%f p:%f\n",dose,p);  
       return p; 
}

void CLogProbit::initializeParameters(){

     CMatrix finalEst;
      const double s_alpha = -5; 
      const double s_beta  = 1.1;
      const double s_gamma = 0.00; 
      double like_val = 0.0, temp_like_val=0.0; 
      double *v = new(double[nparms]);
      double *temp = new(double[nparms]);
      v[LOGPROBIT_ALPHA] = s_alpha;
      v[LOGPROBIT_GAMMA] = s_gamma;
      v[LOGPROBIT_BETA]  = s_beta;
      int dataLength = data.nrow(); 
      double pEst[dataLength],pp[dataLength]; 
      double dose, n, obs; 
      int nonZeroDose = 0; 
      double w_temp; 
      
      memcpy(temp,v,nparms*sizeof(double));
      like_val = temp_like_val; 
      //start the first step of the fisher scoring algorithm
      // using weighted least squared regression
      for (int i = 0; i < dataLength; i++){
          
          n = data.get(i,1);    obs = data.get(i,2);
          dose = data.get(i,0);
          pEst[i] =  (obs+0.5)/(n+1.0);                           
          if (dose > 0) 
             nonZeroDose++;          
      }
      double ww = dataLength-nonZeroDose; 
      ww = 1/ww; 
      double initGamma = 0; 
      for (int i = 0; i < dataLength; i++){
           n = data.get(i,1);    obs = data.get(i,2);
          dose = data.get(i,0);
          if (dose == 0){
                  initGamma += ww*(obs/n); 
          }
      }
      int counter = 0; 
      double nu,der; 
      double vv = log(10); 
      CMatrix z; z.identity(nonZeroDose,1);
      CMatrix x; x.identity(nonZeroDose,2); 
      // Check if we can actually start the iterative procedure
      if (nonZeroDose >= 2){
            double xList[nonZeroDose]; 
            double ztemp;

      
            CMatrix w; w.zero(nonZeroDose,nonZeroDose); 
            // fill up the design matrix
            // with the proper information for the first iteration of 
            // the weighted least squares regression
            
            
            for(int j=0; j< dataLength; j++){
                dose = data.get(j,0);
                n = data.get(j,1); 
                obs = data.get(j,2);
                pEst[j] = (pEst[j]-initGamma)/(1-initGamma); 

                // give a really really small probability                
                // for the estimate
                if (pEst[j] <= 0) pEst[j] = 0.00001; 
                if (obs/n - initGamma > 0){
                
                                pp[j]= (obs/n-initGamma)/(1-initGamma); 
                }else {

                                pp[j] = 0; 
                }
                 
                if (dose > 0){
                         x.set(counter,1,log(dose)); //dose effect
                         x.set(counter,0,1.0); //Intercept
                         nu = gsl_cdf_ugaussian_Qinv(1-pEst[j]); 
                         der = gsl_ran_gaussian_pdf(nu,1.0); 
                         w_temp = n/(pEst[j]*(1-pEst[j]))*der*der;
                         ztemp = nu + (pp[j]-pEst[j])*der; 
                         z.set(counter,0,ztemp); 
                         w.set(counter,counter,w_temp); 
                         counter++;           
                }   
            }          

            CMatrix xwx = x.transpose()*w*x; 
            CMatrix xwz = x.transpose()*w*z; 
            finalEst = xwx.inverse()*xwz;  
            //copy the results as the starting values
            temp[LOGPROBIT_GAMMA] = initGamma; 
            temp[LOGPROBIT_ALPHA] = finalEst.get(0,0);
            temp[LOGPROBIT_BETA]  = finalEst.get(1,0);
      }
      memcpy(model_parameters,temp,nparms*sizeof(double)); 

      delete(temp);
      delete(v);      
}

void CLogProbit::print_output(){
  
     CMatrix m; 
     m = this->varMatrix() ; 
     double temp;
     double bic = ComputeBIC((CBinomial *)this); 
     double aic = ComputeAIC((CBinomial *)this); 
     printf("MODEL: Log-Probit:\n"); 
     printf("------------------------------------------\n"); 
     printf("Parameters\tEstimate\tStdErr\n"); 
     printf("------------------------------------------\n"); 
     temp = sqrt(m.get(2,2)); 
     if (temp == 0){
          printf("GAMMA:\t\t%f\tN/A\n",model_parameters[2]); 
     }else{
          printf("GAMMA:\t\t%1.12f\t%f\n",model_parameters[2],temp); 
     }

     temp = sqrt(m.get(0,0)); 
     if (temp == 0){
          printf("ALPHA:\t\t%f\tN/A\n",model_parameters[0]); 
     }else{
          printf("ALPHA:\t\t%f\t%f\n",model_parameters[0],temp);            
     }     
     
     temp = sqrt(m.get(1,1)); 
     if (temp == 0){
          printf("BETA:\t\t%f\tN/A\n\n",model_parameters[1]); 
     }else{
          printf("BETA:\t\t%f\t%f\n\n",model_parameters[1],temp);            
     }
                
     if (error_status == 0) {
          printf("Optimization Succeeded\n"); 
     } else{
          printf("****WARNING OPTIMIZATION FAILED: Results Suspect****\n"); 
     }  
} 

