#include <gsl/gsl_multimin.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_sf_pow_int.h>
#include <string.h>
#include "CMultistage.h"
#include "run_dmngb.h"

CMultiStage::CMultiStage(){
  set_numParms(3); 
  polyDegree = 2; 
  class_type = multistage; 
} 

// Destructor for the class
CMultiStage::~CMultiStage(){
      // do not need to free up memory as it is already done by 
      // the CBinomial Destructor   
}

//////////////////////////////////////////////////////////////////
// Function: link_function(const double *v, double dose)
// Purpose : Defines a link function in the form of p = g(v,dose) where
//           v is a vector of parameters to the function  and dose is the 
//           specified dose
// Input   :
//          const double *v - vector of parameters to the link function
//              double dose - dose associated with a probability of response
// Output  :
//          return          - the probability of response associated with the link
// Type of link: Multistage
//          p = gamma+(1-gamma)*(1+exp(-(beta1*dose^1+beta2*dose^2 ...)))
// Modified: 11/17/05
///////////////////////////////////////////////////////////////////
double CMultiStage::link_function(const double *v,double dose){
      double gamma =  v[MULTISTAGE_GAMMA]; 
       double p = 0.0; 
       int pos; 
       double beta; 
       double poly; 
       double curval; 
     
      
       poly = 0; 
       for (int i = 1; i < polyDegree + 1; i++){
           beta = v[i];
           poly += beta*gsl_sf_pow_int(dose,i); 
       }
       if (dose == 0){
           p = gamma; 
       }else{
           p = gamma+ (1-gamma)*(1- exp(-poly)); 
       }
   
                    
       return p; 
}

/////////////////////////////////////////////////////////////////////////////
//Function:  max_likelihod -- called to maximize the likelihood for a 
//                          a binomial dose-response likelihood
//		input:
//			iter       - number of iterations
//			gtol       - parameter tolerance
//			relConv    - relative convergence
// Note: This maximization call will be run by most of the routines, the 
//       weibull, multistage, gamma and logistic models need to rescale the
//       doses. 
//Modified 8/25/2006
//////////////////////////////////////////////////////////////////////////////
void CMultiStage::max_likelihod(int iter,double gtol,double relConv){
     int ErrorFlag;
     double *urparm = new (double[nparms]);
     double maxdose = 0.0; 
     double tempdose = maxdose; 
     if (isData){ 
        
        ///////////////////////////////////////////////////
        //scale doses to be 0<=x<=1
        //find max dose; 
        for (int i = 0; i < data.nrow(); i++){
            tempdose = data.get(i,0); 
            if (tempdose > maxdose)
                maxdose = tempdose; 
        }

        for (int i = 0; i < data.nrow(); i++){
            data.set(i,0,data.get(i,0)/maxdose);    
        }      

        //////////////////////////////////////////////////
        initializeParameters(); 
        double scales =  1;  
     
        ErrorFlag = run_dmngb(nparms,model_parameters,lower_bounds, upper_bounds,
                              0,gtol,relConv,iter,10,
			                 (void (*)())binomial_lk,(void(*)())binomial_g,
                             (long int *) this,urparm,NULL,0,&logLikelihood);
        hasBeenMaximized = true; 
        // rescale parameters and doses
        ///////////////////////////////////////////////////
        //resclae dose & the parameters to the origional scale 
        //NOTE: SPECIAL CODE FOR THE MULTISTAGE MODEL
         for (int i = 0; i < data.nrow(); i++){
              data.set(i,0,data.get(i,0)*maxdose);    
         }
         for (int i = 1; i < nparms; i++){          
             model_parameters[i] = model_parameters[i]*pow(1/maxdose,i);
         }   
         //////////////////////////////////////////////////
        
     }
 
     delete urparm;     
}

void CMultiStage::initializeParameters(){
  
     CMatrix y,x;
     CMatrix temp,initParams; 
     CMatrix nonzero;
     double maxDose; 
     bool initFound; 
     double p = 0.0; 
     double obs,n,dose,gamma,tg=0;
     double value,weight=0; 
     
     // find th maximum dose
     maxDose = data.get(0,0); 
     for (int i = 0; i < data.nrow(); i++){
         if (maxDose < data.get(i,0)){
                     maxDose = data.get(i,0); 
         }
     }
     // reasonably estimate the background parameter
     for (int i = 0; i < data.nrow(); i++){
         n = data.get(i,1);
         obs = data.get(i,2);   
         dose = data.get(i,0); 
         weight += (1-dose/maxDose)*(1-dose/maxDose);
         tg += (1-dose/maxDose)*(1-dose/maxDose)*obs/n;         
     }
     
      
     model_parameters[MULTISTAGE_GAMMA] = 0.0816458 ;

     for (int i = 1; i < nparms; i++){
             model_parameters[i] = 0.1;
     }
     model_parameters[2] = 0.1;         
     double v =  log_likelihood(model_parameters); 
}

void CMultiStage::setDegreePoly(int deg){
       bool dataSet = isData; 
       polyDegree = deg; 
       
       set_numParms(polyDegree+1);// add one for the gamma term      
       isData = dataSet; 
}

CBinomial & CMultiStage::operator = (CBinomial &b) {
         // delete from memory the current estimate
    int tempNparms = nparms; 
    CBinomial::operator =(b);  
   
    //*this = b; 
    if( tempNparms ^= nparms) { // this implies there has been a change
        polyDegree = nparms - 1;
    }
      
    return *this; 
         
}
void CMultiStage::print_output(){
     CMatrix m; 
     m = this->varMatrix() ; 
     double temp;
     double bic = ComputeBIC((CBinomial *)this); 
     double aic = ComputeAIC((CBinomial *)this); 
     printf("MODEL: Multistage, %d-degree polynomial:\n",polyDegree); 
     printf("------------------------------------------\n"); 
     printf("Parameters\tEstimate\tStdErr\n"); 
     printf("------------------------------------------\n"); 
     temp = sqrt(m.get(0,0));
     
     if (temp == 0){
          printf("GAMMA:\t\t%e\tN/A\n",model_parameters[0]);      
     }else{
          printf("GAMMA:\t\t%e\t%e\n",model_parameters[0],temp);      
     }
     for (int i = 1; i <  nparms; i++){
         temp = sqrt(m.get(i,i)); 
          if (temp == 0){
               printf("BETA(%d):\t%e\tN/A\n",i,model_parameters[i]); 
          }else{
               printf("BETA(%d):\t%e\t%e\n",i,model_parameters[i],temp);  
          }
     }
     if (error_status == 0) {
          printf("Optimization Succeeded\n"); 
     } else{
          printf("****WARNING OPTIMIZATION FAILED: Results Suspect****\n"); 
     } 
     
} 
