#include <gsl/gsl_multimin.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>
#include "CQQuadratic.h"
#include "run_dmngb.h"

CQQuadraticDR::CQQuadraticDR(){
  set_numParms(2); 
  class_type = quantal_quadratic; 
} 

// Destructor for the class
CQQuadraticDR::~CQQuadraticDR(){
      // do not need to free up memory as it is already done by 
      // the CBinomial Destructor   
}


// Function LinkFunction
// Quantal Linear Link Function for a binomial distribution 
// that is P = gamma + (1-gamma)*(1-exp(-beta*dose*dose))
// NOTE: In order to prevent numerical instabilities e.g. log (0)
// the minimum probability of response is  1e-8 and the maximum
// probability of response is 1-1e-8
double  CQQuadraticDR::link_function(const double *v,double dose){
       double gamma =  v[QQUADRATICDR_GAMMA];
       double beta  =  v[QQUADRATICDR_BETA]; 
       double p = 0.0; 
       int pos; 
       double curval; 
       
       p = gamma + (1-gamma)*(1-exp(-beta*dose*dose));
       // maintain the likelihood function
       if (p< 0) p = 0; 
       if (p > 1) p = 1;

       return p; 
}

void CQQuadraticDR::initializeParameters(){
     //initialize parameters based on a grid search over the range 
     //alpha=1..10 by 1
     //beta =0.25..2.25 by 0.5
     //gamma=0.01..0.21 by 0.5
      
     CMatrix y,x;
     CMatrix temp,initParams; 
     CMatrix nonzero;
     double maxDose; 
     bool initFound; 
     double p = 0.0; 
     double obs,n,dose,gamma,tg=0;
     double value,weight=0;  
     // find th maximum dose

     maxDose = data.get(0,0); 
     for (int i = 0; i < data.nrow(); i++){
         if (maxDose < data.get(i,0)){
                     maxDose = data.get(i,0); 
         }
     }
     for (int i = 0; i < data.nrow(); i++){
         n = data.get(i,1);
         obs = data.get(i,2);   
         dose = data.get(i,0); 
         weight += (1-dose/maxDose)*(1-dose/maxDose);
         tg += (1-dose/maxDose)*(1-dose/maxDose)*obs/n;         
     }
     gamma = tg/weight; 
     // regress on log(1-p) = log(1-gamma)-XBeta
     // create y     

      model_parameters[QQUADRATICDR_GAMMA] = gamma; 
      model_parameters[QQUADRATICDR_BETA]  = 0.1; 
      
}

void CQQuadraticDR::print_output(){
     CMatrix m; 
     m = this->varMatrix(); 
     double temp; 
     double bic = ComputeBIC((CBinomial *)this); 
     double aic = ComputeAIC((CBinomial *)this); 
     printf("MODEL: Quantal Quadratic:\n"); 
     printf("------------------------------------------\n"); 
     printf("Parameters\tEstimate\tStdErr\n"); 
     printf("------------------------------------------\n"); 
     temp = sqrt(m.get(0,0)); 
     if (temp == 0){
          printf("GAMMA:\t\t%f\tN/A\n",model_parameters[0]);      
     }else{
          printf("GAMMA:\t\t%f\t%f\n",model_parameters[0],temp);      
     }
     temp = sqrt(m.get(1,1)); 
     if (temp == 0){
          printf("BETA:\t\t%f\tN/A\n\n",model_parameters[1]); 
     }else{
          printf("BETA:\t\t%f\t%f\n\n",model_parameters[1],temp);  
     }
    
     if (error_status == 0) {
          printf("Optimization Succeeded\n"); 
     } else{
          printf("****WARNING OPTIMIZATION FAILED: Results Suspect****\n"); 
     }            
} 
 
/////////////////////////////////////////////////////////////////////////////
//Function:  max_likelihod -- called to maximize the likelihood for a 
//                          a binomial dose-response likelihood
//		input:
//			iter       - number of iterations
//			gtol       - parameter tolerance
//			relConv    - relative convergence
// Note: This maximization call will be run by most of the routines, the 
//       weibull, multistage, gamma and logistic models need to rescale the
//       doses. 
//Modified 8/25/2006
//////////////////////////////////////////////////////////////////////////////
void CQQuadraticDR::max_likelihod(int iter,double gtol,double relConv){
     int ErrorFlag;
     double *urparm = new (double[nparms]);
     double maxdose = 0.0; 
     double tempdose = maxdose; 
     if (isData){ 
        
        ///////////////////////////////////////////////////
        //scale doses to be 0<=x<=1
        //find max dose; 
        for (int i = 0; i < data.nrow(); i++){
            tempdose = data.get(i,0); 
            if (tempdose > maxdose)
                maxdose = tempdose; 
        }

        for (int i = 0; i < data.nrow(); i++){
            data.set(i,0,data.get(i,0)/maxdose);    
        }      

        //////////////////////////////////////////////////
        initializeParameters(); 
        double scales =  1;  
     
        ErrorFlag = run_dmngb(nparms,model_parameters,lower_bounds, upper_bounds,
                              0,gtol,relConv,iter,10,
			                 (void (*)())binomial_lk,(void(*)())binomial_g,
                             (long int *) this,urparm,NULL,0,&logLikelihood);
        hasBeenMaximized = true; 
        // rescale parameters and doses
        ///////////////////////////////////////////////////
        //resclae dose & the parameters to the origional scale 
        //NOTE: SPECIAL CODE FOR THE GAMMA MODEL
         for (int i = 0; i < data.nrow(); i++){
              data.set(i,0,data.get(i,0)*maxdose);    
         }    
         model_parameters[QQUADRATICDR_BETA] 
                       = model_parameters[QQUADRATICDR_BETA]*pow((1/maxdose),2);            
         //////////////////////////////////////////////////
        
     }
 
     delete urparm;     
}

