////////////////////////////////////////////////////////
//File: BMA.CPP
//Purpose: Conducts all of the model averaging and parametric bootstrapping
//Created: 10/25/2005
//Last Modified:
//       8/25/2006 - Changed a few problems with the selection of the model average
//                   routine.  Bootstrap was computing Added risk regardless of what
//                   risk type was specified, this was a result of the change from 
//                   "simulation code" to "user friendly code". 
  




#include <cstdlib>
#include <iostream>
#include <gsl/gsl_cdf.h>
#include "CBinomial.h"
#include "CQlinearDR.h"
#include "CWeibullDR.h"
#include "CQQuadratic.h"
#include "CLogistic.h"
#include "CProbit.h"
#include "CGamma.h"
#include "CLogLogit.h"
#include "CLogProbit.h"
#include "CMultistage.h"
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>
#include <time.h>

using namespace std;
const int ADDED_RISK = 1;
const int EXTRA_RISK = 2;
const int NUM_DR_MODELS = 9;
extern int randomSeed; 

double computeMAGF(double weights[], CBinomial *models[], const bool incl[],
                   CMatrix data); 
                   
void computeWeights(double weights[], double (*IC)(CBinomial *),const bool incl[],
                    CBinomial *model[]); 
double  compute_acceleration(int riskType, CBinomial *FitModels[], double BMR, 
                           const bool incl[], CMatrix data,
                           int nModels,int degreePoly,
                           double (*IC)(CBinomial *),
                           double *ubounds[],double *lbounds[]);

void  bootstrap_bmd_bmdl(int riskType, CBinomial *FitModels[], double BMR, 
                         const bool incl[], double weights[], 
                         double bmds[], gsl_rng *ranNum, CMatrix data,
                         int bootstraps, int nModels,int degreePoly,
                         double (*IC)(CBinomial *),
                         double *ubounds[],double *lbounds[],double GFBoots[],
			 double *BSparms[]);

inline double predictMA(CBinomial *model[],double dose,double weights[],const bool inclmod[],
                      int nparms); 
                      
double added_risk(CBinomial *model[],double weights[],const bool inclmod[],
                      int nparms,double bmr);
                      
double extra_risk(CBinomial *model[],double weights[],const bool inclmod[],
                      int nparms,double bmr); 

double ComputeBIC(CBinomial *like){
     double parms = (double) like->numParms(); 
     double n = 0.0; 
     //Retrieve the number of observation
     CMatrix data = like->returnData(); 
     for (int i = 0; i < data.nrow();i++){
         n += data.get(i,1);
     }
     return  like->Neg2LogLike()+log(n)*parms; 
}

/*ComputeAIC
  Function: ComputeAIC(CBinomial *like)
  Computes the Corrected AIC of the fit Model
*/ 
double ComputeAIC(CBinomial *like){
     double parms = (double) like->numParms(); 
     return  like->Neg2LogLike()+2*parms; 
}  

double ComputeBICM(CBinomial *like){
     double parms = (double) like->numNBParms(); 
     double n = 0.0; 
     //Retrieve the number of observation
     CMatrix data = like->returnData(); 
     for (int i = 0; i < data.nrow();i++){
         n += data.get(i,1);
     }
     return  like->Neg2LogLike()+log(n)*parms; 
}


/*ComputeAIC
  Function: ComputeAIC(CBinomial *like)
  Computes the Corrected AIC of the fit Model
*/ 
double ComputeAICM(CBinomial *like){
     double parms = (double) like->numNBParms(); 
     return  like->Neg2LogLike()+2*parms; 
}  

double ComputeKIC(CBinomial *like){
     double parms = (double) like->numParms(); 
     return  like->Neg2LogLike()+3*parms; 
}  

double ComputeKICM(CBinomial *like){
     double parms = (double) like->numNBParms(); 
     return  like->Neg2LogLike()+3*parms; 
}  
extern int GLOBAL_numIter;
extern double GLOBAL_relConv; 
extern double GLOBAL_gConv; 
/*Function: AnalyzeData (CBinomial *model, double *ubounds, double *lbounds)
Purpose: Wrapper function to perform a general optimization
*/
void AnalyzeData(CBinomial *model,  double  *ubounds, double *lbounds){
     
     model->setLowerBounds(lbounds);  
     model->setUpperBounds(ubounds);
     model->max_likelihod(GLOBAL_numIter,GLOBAL_relConv,GLOBAL_gConv);   
}

////////////////////////////////////////////////////////
/*Function: modelAverage
  Purpose:  Performs Model Averaging
            dataM - Dichotomous Dose Response data to be fit
            incl  - Pointer to the models which should be included
            ubounds - the upper bounds for the individual DR models
            lbounds - the lower bounds for the individual DR models
            degreePoly - the degree of the multistage polynomial
            riskType   - Choice of Added or Extra Risk
            n_boot     - Number of bootstrap iterations
            bmd        - The estimated bmd
            bmdl       - A vector of length n_boot
            BMR        - Benchmark response to calculate
            IC         - A pointer to the information criterion to be used
  Returns: A modified *bmd and *bmdl vector. The vector bmdl will be in sorted
           descending order.
            
*/
////////////////////////////////////////////////////////////////
void modelAverage(CMatrix dataM, const bool *incl, double *ubounds[], 
                  double *lbounds[], int degreePoly, int riskType,  
                  int n_boot, double *bmd, double *bmdl, 
                  double BMR, double (*IC)(CBinomial *),double *a, 
                  double *weights, CBinomial *cpyModels[],double *GF_test,
                  double *GF_P_value, double *BSparms[], int *maxBSparms){

     // set up the Likelihoods for the models
     CQlinearDR         qlin;     qlin.setData(dataM); 
     CWeibullDR      weibull;     weibull.setData(dataM); 
     CQQuadraticDR     qquad;     qquad.setData(dataM); 
     CLogistic      logistic;     logistic.setData(dataM); 
     CProbit          probit;     probit.setData(dataM); 
     CMultiStage      mstage;     mstage.setData(dataM);    
     CLogProbit      logProbit;   logProbit.setData(dataM); 
     CLogLogistic    logLogistic; logLogistic.setData(dataM);
     CGamma          gamma;       gamma.setData(dataM); 
     mstage.setDegreePoly(degreePoly);
     
     CMultiStage  test; 
     
     // set the bounds for the different models
    
     
     CBinomial *models[] = {&qlin,&qquad,&mstage,&logistic,
                             &probit,&weibull,&logProbit,&logLogistic,
                             &gamma};
     *maxBSparms = 0;
    
     
     // fit the models that have been specified                     
     for (int i = 0; i < NUM_DR_MODELS; i++){
         if(incl[i]){
             AnalyzeData(models[i], ubounds[i],lbounds[i]);
	     // Count up the number of columns to be used in BSparms
	     *maxBSparms += 1 + models[i]->get_numParms();
         }     
     }

     
     // get weights which are formed from the given information 
     // criterion    
     computeWeights(weights,IC,incl,models);      
     
     //*estimate the benchmark dose
     double xx; 
     if (riskType == ADDED_RISK){  
          xx = added_risk(models,weights,incl,NUM_DR_MODELS,BMR); 
     }else{     
          xx = extra_risk(models,weights,incl,NUM_DR_MODELS,BMR); 
     }     
     *bmd = xx;
     
     double gfValue = computeMAGF(weights, models, incl, dataM);
     double GFBoots[n_boot]; 
     
     // Set up the random number generator for the bootstrap
     gsl_rng *ranNum = gsl_rng_alloc(gsl_rng_mt19937); 
     gsl_rng_set(ranNum,randomSeed); 
     

     bootstrap_bmd_bmdl(riskType, models, BMR, 
                        incl, weights, bmdl, ranNum, dataM,
                        n_boot, NUM_DR_MODELS, degreePoly,
                        IC,ubounds,lbounds,GFBoots,BSparms);

     ////////////////////////////////////////////////////////////
     //sort the bootstrap in a new vector and copy it back in    
     gsl_vector *temp = gsl_vector_alloc(n_boot); 
     for (int j = 0; j < n_boot; j++){
         gsl_vector_set(temp,j,bmdl[j]); 
     }
     gsl_sort_vector(temp); 
 
     for (int j = 0; j < n_boot; j++){
         bmdl[j] = gsl_vector_get(temp,j); 
     }
     gsl_vector_free(temp);
     ////////////////////////////////////////////////////////////  
     //sort the bootstrap GF Test in    
     temp = gsl_vector_alloc(n_boot); 
     for (int j = 0; j < n_boot; j++){
         gsl_vector_set(temp,j,GFBoots[j]); 
     }
     gsl_sort_vector(temp); 
     int GFpos = 0; 
     for (int j = 0; j < n_boot; j++){
         GFBoots[j] = gsl_vector_get(temp,j); 
         if (gfValue >= GFBoots[j]){
                     GFpos = j;    }      
     }
     double pval = 1.0 - (double(GFpos)/double(n_boot)); 
  
     *GF_test = gfValue; 
     *GF_P_value = pval; 
     gsl_vector_free(temp);
     //////////////////////////////////////////////////////////////
 
     double acc = compute_acceleration(riskType, models, BMR,
                                      incl, dataM,
                                      NUM_DR_MODELS, degreePoly,
                                      IC,ubounds,lbounds);
                                                                   
     *a = acc; 

     for (int i = 0; i < NUM_DR_MODELS; i++){
         *cpyModels[i] = *models[i];
         
     }       
}

/*predictMA
*model   : the model vector
dose     : the dose to predict 
*weights : the weigts of individual models
*inclmod : a boolean vector 
Purpose  :  Produces a weigted average 
*/
inline double predictMA(CBinomial *model[],double dose,double weights[],const bool inclmod[],
                      int nparms){
       double predVal = 0.0; 
       for (int i = 0 ; i < nparms; i++){
           if (inclmod[i]){
	     predVal += model[i]->predict(dose)*weights[i];
           }
       }       
       return predVal; 
} 
/////////////////////////////////////////////////
// Function: added_risk
// input
//      -CBinomial models[]: A vector of binomial classes that are fit
//      -double weights[]:   A vector of weights that correspond to the model
//      -const bool inclmod[]: The models that are included in the analysis
//      -int nparms:          the number of parameters in the vector(s)
//      -double bmr:          The specified BMR
// output: (double) estimated Benchmark dose corresponding to the MA
//          computed using the method of bisection.  Note: 10e6 is returned
//          if the function can't bound the BMD. 
///////////////////////////////////////////////////
double added_risk(CBinomial *model[],double weights[],const bool inclmod[],
                      int nparms,double bmr){
     double background = predictMA(model,0.0,weights,inclmod,nparms); 
     
     //attempt to find a dose such that pi(dose)-pi(background) >= bmr
     // if it still is not good after 20 iterations then we break out out of 
     // the function returning a BMD of 2^20 = 1e6 
     double tempDose = 0.5; 
     int   multiplier  = 1; 
     double diff;  
     bool bounded_bma = false; 
     
     for (int i = 1; i <=20; i++){
         multiplier *= 2; 
         tempDose = 0.5*multiplier; 
     
         diff = predictMA(model,tempDose,weights,inclmod,nparms)-background; 
         if (diff >= bmr) 
         {   bounded_bma  = true; 
             break;          
         }
     }

     if (!bounded_bma){
              return (double)1e6;     
     }
     
     // Some changes to search on log(BMD) instead of BMD.
     // double top = tempDose; 
     double top=log(tempDose);
     // double bottom = 0; 
     double bottom = log(2.2e-308);
     double mid = (top+bottom)/2; 
     int iterations = 1; 
     
     // First, make sure the BMD is above bottom (we already know it is below top)
     // if BMD is below bottom, just return 0, as bottom is already as small as
     // we can represent.
     double td=exp(bottom);
     diff = (predictMA(model,td,weights,inclmod,nparms) - background); 
     if (diff > bmr) {
       return 0.0;
     } else {
       while ( fabs(bmr-diff) >= 1.0e-10 && fabs(top - bottom) >= 1.0e-10 &&
	       iterations <=400){
	 tempDose = exp(mid); 
	 diff =  predictMA(model,tempDose,weights,inclmod,nparms) - background; 
	 
	 if (diff > bmr)
	   top = mid; 
	 else 
	   bottom = mid; 
	 
	 mid = (top+bottom)/2; 
	 iterations ++; 
       }
       
       if (iterations > 400) // if there is a problem
	 return -1.0;       // return a -1.0 dose
       else
         return tempDose; 
     }
}
/////////////////////////////////////////////////
// Function: extra_risk
// input
//      -CBinomial models[]: A vector of binomial classes that are fit
//      -double weights[]:   A vector of weights that correspond to the model
//      -const bool inclmod[]: The models that are included in the analysis
//      -int nparms:          the number of parameters in the vector(s)
//      -double bmr:          The specified BMR
// output: (double) estimated Benchmark dose corresponding to the MA
//          computed using the method of bisection.  Note: 10e6 is returned
//          if the function can't bound the BMD. 
///////////////////////////////////////////////////
double extra_risk(CBinomial *model[],double weights[],const bool inclmod[],
                      int nparms,double bmr){
     double background = predictMA(model,0.0,weights,inclmod,nparms); 
     
     //attempt to find a dose such that pi(dose)-pi(background) >= bmr
     // if it still is not good after 20 iterations then we break out out of 
     // the function returning a BMD of 2^20 = 1e6 
     double tempDose = 0.5; 
     int   multiplier  = 1; 
     double diff;  
     bool bounded_bma = false; 
     for (int i = 1; i <=20; i++){
         multiplier *= 2; 
         tempDose = 0.5*multiplier; 
         diff = (predictMA(model,tempDose,weights,inclmod,nparms)-background)/(1-background); 
         if (diff >= bmr) 
         {   bounded_bma  = true; 
             break;          
         }
     }
     
     if (!bounded_bma){
              return (double)1e6;     
     }
     // Some changes to search on log(BMD) instead of BMD.
     //double top = tempDose; 
     double top=log(tempDose);
     //double bottom = 0; 
     double bottom = log(2.2e-308);
     double mid = (top+bottom)/2; 
     int iterations = 1; 
     
     // First, make sure the BMD is above bottom (we already know it is below top)
     // if BMD is below bottom, just return 0, as bottom is already as small as
     // we can represent.
     double td=exp(bottom);
     diff = (predictMA(model,td,weights,inclmod,nparms) - background)/(1-background); 
     if (diff > bmr) {
       return 0.0;
     } else {
       while ( fabs(bmr-diff) >= 1.0e-10 && fabs(top - bottom) >= 1.0e-10 &&
	       iterations <=400){
	 tempDose = exp(mid); 
	 diff =  (predictMA(model,tempDose,weights,inclmod,nparms) - background)/(1-background); 
	 
	 if (diff > bmr)
	   top = mid; 
	 else 
	   bottom = mid; 
	 
	 mid = (top+bottom)/2; 
	 iterations ++; 
       }
       
       if (iterations > 400) // if there is a problem
	 return -1.0;       // return a -1.0 dose
       else
	 return tempDose; 
     }
}
////////////////////////////////////////////////////////////////////////////
// Function : bootstrap_bmd_bmdl()
// purpose:  Computes a number of parametric bootstrap resamples of the MA
// input:  
//         riskType- The type of risk the bootstrap is calculating
//         CBinomial FitModels[] - The models that were fit, containing the MLE
//         const bool incl[]     - The models included in the analysis
//         double weights[]      - the MA weights
//         double bmds[]         - a vector that has the bootstrap resamples
//         gsl_rng ranNum        - The random number seed
//         CMatrix data          - The matrix of data
//         int bootstraps        - the number of bootstraps to perform
//         int nModels           - the number of models in the FitModels[] vector
//         int degreePoly        - The degree of the multistage polynomial
//         double (*IC) (CBinomial*) - Callback function containing the correct
//                                   information criterion to use in the MA
//         double ubounds        - the upper bounds for all of the models
//         double lbounds        - the lower bounds for all of the models
// output: the vector bmds[] will be modified containing (int bootstraps) number
//         of bootstraps.
//         the array BSparms[][] will be modified to contain the weights and parameter
//         values for the bootstrap fits.
//         
/////////////////////////////////////////////////////////////////////////////   
void  bootstrap_bmd_bmdl(int riskType, CBinomial *FitModels[], double BMR, 
                         const bool incl[], double weights[], 
                         double bmds[], gsl_rng *ranNum, CMatrix data,
                         int bootstraps, int nModels,int degreePoly,
                         double (*IC)(CBinomial *),
                         double *ubounds[],double *lbounds[],double GFBoots[],
			             double *BSparms[]){
                             
     // set up the Likelihoods for the bootstrap
     CQlinearDR         qlin;    
     CWeibullDR      weibull;     
     CQQuadraticDR     qquad;     
     CLogistic      logistic;     
     CProbit          probit;     
     CMultiStage      mstage;        
     CLogProbit      logProbit;   
     CLogLogistic    logLogistic; 
     CGamma          gamma;      
     mstage.setDegreePoly(degreePoly);
             
     CBinomial *models[] = {&qlin,&qquad,&mstage,&logistic,
                             &probit,&weibull,&logProbit,&logLogistic,
                             &gamma};
                             
     
     CMatrix bsData = data; 
     double obs,dose,n,p; 
    
       
     for (int k = 0; k < bootstraps; k++){
        // perform the parametric bootstrap by creating new data
	
      
         for (int j = 0; j < bsData.nrow(); j++){
              dose = data.get(j,0);
              n = data.get(j,1);
              p =  predictMA(FitModels,dose,weights,incl,nModels);                 
              obs = (double) gsl_ran_binomial (ranNum, p, (unsigned int)n);            
              bsData.set(j,2,obs);  
	                     
         }
     
         double ic[NUM_DR_MODELS]; 
         double weights[NUM_DR_MODELS];
         double sum_ic = 0;

         // fit the models that have been specified                     
         for (int i = 0; i < NUM_DR_MODELS; i++){
              if(incl[i]){
                    models[i]->setData(bsData);     
                    AnalyzeData(models[i], ubounds[i],lbounds[i]);    
              }     
         }

        //Calculate weights for the new bootstrap resample
         computeWeights(weights,IC,incl,models);  
	 //Copy the weights and parameters to kth row of BSparms;
	 int j = 0;
	 for (int i = 0; i < NUM_DR_MODELS; i++) {
	   if (incl[i]) {
	     BSparms[k][j] = weights[i]; j++;
	     models[i]->copy_Parms(&BSparms[k][j]);
	     j += models[i]->get_numParms();
	     }
	 }
	     
	       
         GFBoots[k] = computeMAGF(weights, models, incl, bsData);
         // Calculate the bootstrap BMD estimate  
         if (riskType == ADDED_RISK){  
              bmds[k] = added_risk(models,weights,incl,NUM_DR_MODELS,BMR); 
         }else{     
              bmds[k] = extra_risk(models,weights,incl,NUM_DR_MODELS,BMR); 
         }
              
     }
       
}
////////////////////////////////////////////////////////////////////
// Computes the acceleration as in "Introduction to the Bootstrap"
// B. Efron and R.J.Tibshirani. 
// NOTE: The jacknife is computed effieciently in that it weights the removal of 
// a specific observation by the number of successes, or failures.  Thus there are
// only 2*(number of data lines) calculations. 
////////////////////////////////////////////////////////////////////
double  compute_acceleration(int riskType, CBinomial *FitModels[], double BMR, 
                           const bool incl[], CMatrix data,
                           int nModels,int degreePoly,
                           double (*IC)(CBinomial *),
                           double *ubounds[],double *lbounds[]){
     double acceleration = 0;                                   
     
                                  
 // set up the Likelihoods for the models
     CQlinearDR         qlin;    
     CWeibullDR      weibull;     
     CQQuadraticDR     qquad;     
     CLogistic      logistic;     
     CProbit          probit;     
     CMultiStage      mstage;        
     CLogProbit      logProbit;   
     CLogLogistic    logLogistic; 
     CGamma          gamma;      
     mstage.setDegreePoly(degreePoly);
     // set the bounds for the different models
    
     
     CBinomial *models[] = {&qlin,&qquad,&mstage,&logistic,
                             &probit,&weibull,&logProbit,&logLogistic,
                             &gamma};
                             
    
     CMatrix bsData = data; 
     double obs,dose,n,p; 
     int size = bsData.nrow();
     double temp; 
     
     // the following code performs a jacknife on the BMD estimate
     double observed_bmd[size],non_observed_bmd[size];    
     for (int k = 0; k < data.nrow(); k++){
         // look at a success or failure
         for (int j = 0 ; j < 2; j++){
              bsData = data; 
              dose = data.get(k,0);
              n = data.get(k,1)-1;
              obs = data.get(k,2);
              if ( j == 1){ // if  I am removing an observed positive response
                 obs --; 
                 if (obs < 0) obs = 0;                    
              }else{ // otehrwise I am doing nothing
                    
              }
              if (n == 0){
                 bsData.deleterow(k);      
              }else{
                  bsData.set(k,2,obs);    
                  bsData.set(k,1,n); 
              }
                
              double ic[NUM_DR_MODELS]; 
              double weights[NUM_DR_MODELS];
              double sum_ic = 0;
              // fit the models that have been specified                     
         
              for (int i = 0; i < NUM_DR_MODELS; i++){
                 if(incl[i]){
                       models[i]->setData(bsData);     
                       AnalyzeData(models[i], ubounds[i],lbounds[i]);    
                 }     
              }
         
              for (int i = 0; i < NUM_DR_MODELS; i++){
                   ic[i] = 0; //initialize
                   if(incl[i]){
                        ic[i] = IC(models[i]);
                        sum_ic += exp(-0.5*ic[i]); 
                   }    
             }
     
             // Compute the weights
             for (int i = 0; i < NUM_DR_MODELS; i++){
                  weights[i] = 0; 
                  if(incl[i]){
                       weights[i] = exp(-0.5*ic[i])/sum_ic;
                  }
             }
           
             if (riskType == ADDED_RISK){  
                 temp = added_risk(models,weights,incl,NUM_DR_MODELS,BMR); 
             }else{     
                 temp = extra_risk(models,weights,incl,NUM_DR_MODELS,BMR); 
             }
             
             if (j == 1){
                   observed_bmd[k] = temp; 
             }else{
                   non_observed_bmd[k] = temp; 
             }
                   
         }
     }   
     

       
     // calculate n
     double jack_bmd = 0; 
     double nn = 0; 
     for (int i = 0; i < data.nrow(); i++){
         nn +=  data.get(i,1);
     }
     
     // estimate the jacknife 
     for (int i = 0; i <  data.nrow(); i++){
          for (int j = 0; j < 2; j++){
              n = data.get(i,1);
              obs = data.get(i,2);
              if (j == 0) { // weight the sample by how many there are
                  jack_bmd += (n - obs)*non_observed_bmd[i]/nn; 
              }else {
                  jack_bmd +=  obs*observed_bmd[i]/nn;   
              }              
          }    
     }             
         
     
     // estimate the acceleration
     double  sum_num = 0;
     double sum_denom = 0; 
     for (int i = 0; i <  data.nrow(); i++){
          for (int j = 0; j < 2; j++){
              n = data.get(i,1);
              obs = data.get(i,2);
              if (j == 0) { // weight the sample by how many there are
                  sum_num   += (n - obs)*pow(jack_bmd - non_observed_bmd[i],3.0); 
                  sum_denom += (n - obs)*pow(jack_bmd - non_observed_bmd[i],2.0); 
              }else {
                  sum_num   += (obs)*pow(jack_bmd - observed_bmd[i],3.0);
                  sum_denom += (obs)*pow(jack_bmd - observed_bmd[i],2.0);   
              }              
          }    
     }                     
     acceleration = sum_num/(6.0*pow(sum_denom,1.5)); 
     return acceleration; 
                                  
}
//////////////////////////////////////////////////////////////////
//Purpose: Computes the model average weights given a weighting criterion
//Input: double (*IC) (CBinomial *)  -- A pointer to a function that represents
//                                      an information criterion
//       bool  incl[]                  --  A vector specifying if the model 
//                                       is included in the model average
//Output double weights[]            -- A vector of weights
////////////////////////////////////////////////////////////////////////
void computeWeights(double weights[], double (*IC)(CBinomial *), const bool incl[],
                    CBinomial *models[])
{
     double ic[NUM_DR_MODELS];    
     double sum_ic = 0;
     
     for (int i = 0; i < NUM_DR_MODELS; i++){
         ic[i] = 0; //initialize
         if(incl[i]){
             ic[i] = IC(models[i]);
             sum_ic += exp(-0.5*ic[i]); 
         }    
     }
     
     // Compute the weights
     for (int i = 0; i < NUM_DR_MODELS; i++){
         weights[i] = 0; 
         if(incl[i]){
             weights[i] = exp(-0.5*ic[i])/sum_ic;
         }
       //  printf("Iteration:%d,weights:%f\n",i,weights[i]);         
     }
}
////////////////////////////////////////////////////////////////////////////
//Purpose: Computes the Model Average Goodness of fit statistic
//Input:    double weights[]   -- MA Weights
//          CBinomial *models[]-- The fitted model average models
//          const bool incl[]  -- The models that are included in the MA
//Output:   The calculated GF statistic
///////////////////////////////////////////////////////////////////////////
double computeMAGF(double weights[], CBinomial *models[], const bool incl[],
                   CMatrix data){
         double dose = 0; double n = 0; double p = 0; double obs; 
         double gfValue = 0.0; 
         for (int j = 0; j < data.nrow(); j++){
              dose = data.get(j,0);
              n   = data.get(j,1);
              obs = data.get(j,2);
              p =  predictMA(models,dose,weights,incl,NUM_DR_MODELS); 
              if ( p == 0){ p = 1e-8;} 
              gfValue += (obs/n-p)*(obs/n-p)/(p*(1-p)/n);                                                      
         }
         
         return gfValue;       
}
